/********** Modified from RelativeEncodingGenome.java ************/

import java.text.DecimalFormat;
import java.util.*;


class AbsoluteEncodingGenome  extends Genome {
	static AbsoluteEncodingLscape  lmls;
	static char	alphabet[] = {'F', 'B', 'L', 'R', 'U', 'D'};

	char    gene[];

	static int        mutation_step_size = 1;
	static boolean    adaptive_mutation = false;
	static float      adaptive_step_size = 0.1F;
	static float      mutation_success_rate;

	/*
   Genome constructor: uses a uniform random number generator to assign
                initial values to each gene from the range specified 
		by the the given lower and upper bounds.
	 */

	AbsoluteEncodingGenome(EC_rand rand){
		gene = new char[gene_cnt];
		for (int i=0; i<gene_cnt; i++)   gene[i] = alphabet[rand.irand(alphabet.length)];
	}

	AbsoluteEncodingGenome(){
		gene = new char[gene_cnt];
	}
	AbsoluteEncodingGenome(char[] a){
		gene = a;
	}
	
	public static int tempCount=0;
	void run_reset(EC_rand rand){
//		gene[0] = alphabet[rand.irand(alphabet.length)];
//		do{
//			for (int i=1; i<gene_cnt; i++){
//				char a;
//				do{
//					a = alphabet[rand.irand(alphabet.length)];
//				}while(!oneStepAvoiding(gene[i-1], a));
//				gene[i] = a;
//			}
//			
//		}while(!this.legal());
		for (int i=0; i<gene_cnt; i++)   gene[i] = 'F';
//		do{
//			for (int i=0; i<gene_cnt; i++)   gene[i] = alphabet[rand.irand(alphabet.length)];
//		}while(!this.legal());
//		System.out.println(tempCount++);
//		System.out.println("****************escaped!****************");
		
	}
	
	boolean oneStepAvoiding(char a, char b){
		if( (a=='F' && b=='B') || (a=='B' && b=='F')){
			return false;
		}else if ( (a=='R' && b=='L') || (a=='L' && b=='R')){
			return false;
		}else if ( (a=='U' && b=='D') || (a=='D' && b=='U')){
			return false;
		}else{
			return true;
		}
	}

	/*
   IntegerGenomeClassInit: initializes the genome/phenome mapping for entire class
	 */

	static void AbsoluteEncodingGenomeClassInit(Properties props, Landscape lscape, EA ea) 
	{

		if (!lscape.lscape_type.equals("AbsoluteEncoding"))
			Util.ec_error("Can't map " + lscape.lscape_type + " into absolute encoding representation");

		lmls = (AbsoluteEncodingLscape) lscape;

		gene_cnt = lmls.param_cnt;

		System.out.println("  Using a genome with " + gene_cnt + " absolute encoding(char) gene(s).");

		genotype_cardinality = lmls.genotype_cardinality;

		ea.global_max = new AbsoluteEncodingIndividual();
		ea.global_min = new AbsoluteEncodingIndividual();
		ea.global_max.fitness = Float.NEGATIVE_INFINITY;
		ea.global_min.fitness = Float.POSITIVE_INFINITY;
		ea.parent_pop = new AbsoluteEncodingPopulation(ea, ea.parent_pop_size, ea.rand);

		GenomeClassInit(props, ea);

		System.out.println("Absolute Encoding: Mutation type using default.");
		String mutation_spec = "random";
		mutation_type = random_mutation;
		mutation_probability = 1.0F / Genome.gene_cnt;

		String mutation_options = props.getProperty("Mutation");
		if(mutation_options != null){
			StringTokenizer st = new StringTokenizer(mutation_options," ");
			st.nextToken();
			if(st.hasMoreTokens()){
				mutation_probability = Float.parseFloat(st.nextToken());
			}
		}

	}




	/*
  mutate:  mutates the given genome
	 */

	void mutate(EA ea)
	{
		switch (mutation_type) {

		case random_mutation:
			random_mutate(ea.rand);
			break;
		default:
			Util.ec_error("illegal Absolute Encoding mutation type: " + mutation_type);
		}
	}

	void random_mutate(EC_rand rand)
	{
		char new_val[] = new char[gene_cnt];
		int mutated = 0;

		for (int i=0; i<gene_cnt; i++)
		{
			if (rand.frand() <= mutation_probability)
			{     
				new_val[i] = alphabet[rand.irand(alphabet.length)];
				mutation_step_sum += 1;
				mutated++;
			}
			else  new_val[i] = gene[i];
		}

		if (mutated>0)
		{
			if (lmls.legal(new_val))
				gene = new_val;
			else
				invalid_mutation_count++;
		}

		mutation_count += mutated;
		return;
	}
	
	


	/*
  crossover:
	 */

	void crossover(Genome mate, EC_rand rand)
	{
//		int count=0;
//		do{
			switch (crossover_type) {
	
			case one_point:
				one_pt_crossover((AbsoluteEncodingGenome)mate, rand);
				break;
	
			case two_point:
				two_pt_crossover((AbsoluteEncodingGenome)mate, rand);
				break;
	
			case uniform:
				uniform_crossover((AbsoluteEncodingGenome)mate, rand);
				break;
	
			default:
				Util.ec_error("illegal Integer crossover type: " + crossover_type);
			}
//			count++;
//			if(count>1000){
//				System.err.println("Oh, my god. Forget it");
//				break;
//			}
//		}while( ! lmls.legal(this.gene));
	}


	/*
   one_pt_crossover:  copies selected genes from mate into this genome
	 */

	void one_pt_crossover(AbsoluteEncodingGenome mate, EC_rand rand)
	{
		int cut_point, i;
		char tmp;

		if (gene_cnt == 1)  return;    // need at least two genes

		if (keep_both_sides) {
			if (gene_cnt == 2) {	//easy case
				tmp = this.gene[1];
				this.gene[1] = mate.gene[1];
				mate.gene[1] = tmp;
			}
			else {
				cut_point = rand.irand(gene_cnt-1)+1;
				for (i=cut_point; i<gene_cnt; i++)  {
					tmp = this.gene[i];
					this.gene[i] = mate.gene[i];
					mate.gene[i] = tmp;
				}
			}
		}
		else {
			if (gene_cnt == 2)  this.gene[1] = mate.gene[1];   // easy case
			else {
				cut_point = rand.irand(gene_cnt-2);
				for (i=cut_point+1; i<gene_cnt; i++)  this.gene[i] = mate.gene[i];
			}
		}

		return;
	}


	/*
   two_pt_crossover:  copies selected genes from mate into this genome
	 */

	void two_pt_crossover(AbsoluteEncodingGenome mate, EC_rand rand)
	{
		int cut_point1, cut_point2, i;
		char tmp;

		if ((gene_cnt == 1) || (gene_cnt ==2))  return;  // need at least 3 genes

		if (keep_both_sides) {
			if (gene_cnt == 3) {  // easy case
				tmp = this.gene[1];
				this.gene[1] = mate.gene[1];
				mate.gene[1] = tmp;
			}
			else
			{
				cut_point1 = rand.irand(gene_cnt);
				cut_point2 = rand.irand(gene_cnt);
				while (cut_point2 == cut_point1) 
					cut_point2 = rand.irand(gene_cnt);  // want distinct cut points

				if (cut_point1 < cut_point2)
					for (i=cut_point1+1; i<=cut_point2; i++)  {
						tmp = this.gene[i];
						this.gene[i] = mate.gene[i];
						mate.gene[i] = tmp;
					}
				else
					for (i=cut_point2+1; i<=cut_point1; i++)  {
						tmp = this.gene[i];
						this.gene[i] = mate.gene[i];
						mate.gene[i] = tmp;
					}
			}
		}
		else {
			if (gene_cnt == 3)  this.gene[1] = mate.gene[1];  // easy case
			else
			{
				cut_point1 = rand.irand(gene_cnt);
				cut_point2 = rand.irand(gene_cnt);
				while (cut_point2 == cut_point1) 
					cut_point2 = rand.irand(gene_cnt);  // want distinct cut points

				if (cut_point1 < cut_point2)
					for (i=cut_point1+1; i<=cut_point2; i++)  this.gene[i] = mate.gene[i];
				else
					for (i=cut_point2+1; i<=cut_point1; i++)  this.gene[i] = mate.gene[i];
			}
		}

		return;
	}






	/*
   uniform_crossover:  copies genes from mate into this genome probabilistically
	 */

	void uniform_crossover(AbsoluteEncodingGenome mate, EC_rand rand)
	{
		int i; char tmp;

		if (keep_both_sides) {
			for (i=0; i<gene_cnt; i++)  
				if (rand.frand() <= uniform_crossover_probability) {
					tmp = this.gene[i];  
					this.gene[i] = mate.gene[i];
					mate.gene[i] = tmp;
				}
		}
		else {
			for (i=0; i<gene_cnt; i++)  
				if (rand.frand() <= uniform_crossover_probability)  
					this.gene[i] = mate.gene[i];
		}

		return;
	}



	/*
   global_stats:  print global operator stats
	 */

	void global_stats()
	{
		System.out.println(" Total number of mutations: " + mutation_count);
		System.out.println(" Number of constraint violations: " + 
				invalid_mutation_count);
		System.out.println(" Average mutation step size: " +
				mutation_step_sum/mutation_count);
	}



	/*
   eval:
	 */

	float eval()
	{
		return (lmls.eval(gene));
	}




	DecimalFormat i6 = new DecimalFormat("######0");


	/*
   print_genome:  prints the gene values as real numbers
	 */

	void print()
	{
		int  i;

		for (i=0; i<gene_cnt; i++)
			System.out.print(gene[i]);
	}



	Genome copy()
	{
		AbsoluteEncodingGenome  dupl = new AbsoluteEncodingGenome();

		dupl.gene_cnt = gene_cnt;
		dupl.gene = new char[gene_cnt];

		System.arraycopy(gene, 0, dupl.gene, 0, gene_cnt);

		return ((Genome) dupl);
	}



	/*
  distance:  computes Hamming distance between self and specified partner
	 */

	float distance(Genome partner)
	{
		int diff=0;

		for (int i=0; i<Genome.gene_cnt; i++) { 
			if(this.gene[i] != ((AbsoluteEncodingGenome)partner).gene[i]){
				diff++;
			}
		}

		return(diff);
	}
	
	@Override
	public AbsoluteEncodingGenome createMirrorGenome(){
		char[] newGene = new char[gene.length];
		for(int i=0; i<gene.length; i++){
			switch(gene[i]){
				case 'F':
					newGene[i] = 'B';
					break;
				case 'B':
					newGene[i] = 'F';
					break;
				case 'R':
					newGene[i] = 'R';
					break;
				case 'L':
					newGene[i] = 'L';
					break;
				case 'U':
					newGene[i] = 'D';
					break;
				case 'D':
					newGene[i] = 'U';
					break;
				default:
					System.err.println("not FRLUD");
					System.exit(-1);
			}
		}
		return new AbsoluteEncodingGenome(newGene);
	}

	public boolean legal(){
		return lmls.legal(gene);
	}
	
	public float avgContactSeperation(){
		return (lmls.avgContactSeperation(gene));
	}
	
	public float volume(){
		return (lmls.volume(gene));
	}
	
	public ArrayList<int[]> contactSeperationSignature(boolean onlyHH){
		return (lmls.contactSeperationSignature(gene, onlyHH));
	}
}





