/*       Copyrighted by Ricky Wong - 2009          */
/* /

/*
    CrowdingGA: Crowding Genetic Algorithm
*/

import java.util.*;


class CrowdingGA   extends EA
{
/*
    CrowdingGA constructor
*/
	int crowdingFactor;
	CrowdingGA(Properties props, Landscape lscape, int rand_seed){
	    super(rand_seed);
	    ea_type = "CrowdingGA";
	    EA_init(props, lscape);         // finish general EA initialization;
	    
	    StringTokenizer st = new StringTokenizer(props.getProperty("EA")," ");
	    st.nextToken();
	    if(st.hasMoreTokens()){
	    	crowdingFactor = Integer.parseInt(st.nextToken());
	    }else{
	    	crowdingFactor = parent_pop_size;
	    }
	    
	    System.out.println("CrowdingGA: It uses Crowding Factor:" + crowdingFactor);
	    
	    
    }
	/*
	   run: the specified EA for the specified time
	*/
	
	void run(int birth_limit)
	{
		birth_cnt = 0;
		generation_cnt = 0;
		NFE = 0;
		global_max.fitness = Float.NEGATIVE_INFINITY;
		global_min.fitness = Float.POSITIVE_INFINITY;
		parent_pop.run_reset();
		dump();
		
		if(crowdingFactor==0){
			Util.ec_error("Crowding Factor == 0 !"); System.exit(-1);
		}
		while ( (birth_cnt < birth_limit) & (NFE <= EC4.maxNFE) ) 
		{
			if (report_fixed_gen || report_fixed_fit)  if (parent_pop.fixed_point()) {
				break;
			}

			offspring_pop = select_and_clone_parents();

			introduce_offspring_variation();

			evaluate_offspring();

			/***************************************************/
			/***********Crowding Selection**********************/
			/***************************************************/
			Individual[] temp = new Individual[parent_pop.pop_size];
			float min_fitness = Float.MAX_VALUE;
			for(int k=0; k<parent_pop.pop_size; k++){
				temp[k] = parent_pop.indiv[k];
				if(min_fitness>temp[k].fitness){
					min_fitness = temp[k].fitness;
				}
			}

			for(int i=0; i<offspring_pop.indiv.length; i++){
				if(offspring_pop.indiv[i].fitness>=min_fitness){
					int NN = nearestNeighborInt(offspring_pop.indiv[i].genome, temp, crowdingFactor);	//**it uses parent_pop
					if(offspring_pop.indiv[i].fitness>temp[NN].fitness){
						temp[NN] = offspring_pop.indiv[i];
					}
				}
			}
			parent_pop.indiv = temp;
			/*******************END******************************/

			generation_cnt ++;
			dump();
//			adapt_operators();
//			adaptation_stats();
		}
		
	}
	
	int nearestNeighborInt(Genome target, Individual[] indivArray){
		int NN = 0;
		double minDistance = target.distance(indivArray[0].genome);
		for(int i=1; i<indivArray.length; i++){
			double temp = target.distance(indivArray[i].genome);
			if(temp<minDistance){
				minDistance = temp; 
				NN=i;
			}
		}
		return NN;
	}
	
	int nearestNeighborInt(Genome target, Individual[] indivArray, int crowdingFactor){
		
		if(crowdingFactor>indivArray.length){
			System.err.println("Ricky: crowdingFactor>indivArray.length");
			System.exit(-1);
		}
		
		/*****************************************/
		// Prepare the candidate individual index
		/*****************************************/
		int candidateIndex[] = generateCandidateIndex(indivArray,crowdingFactor);
		
		/*************************************************/
		// Choose the closest candidate individual index
		/*************************************************/
		int NN = 0;
		double minDistance = target.distance(indivArray[candidateIndex[0]].genome);
		for(int i=1; i<candidateIndex.length; i++){
			double temp = target.distance(indivArray[candidateIndex[i]].genome);
			if(temp<minDistance){
				minDistance = temp; 
				NN=candidateIndex[i];
			}
		}
		return NN;
	}
	
	protected int[] generateCandidateIndex(Individual[] indivArray, int crowdingFactor){
		int indivIndex[] = new int[indivArray.length];
		for(int i=0; i<indivIndex.length; i++){
			indivIndex[i] = i;
		}
		int currentLength = indivIndex.length;
		int candidateIndex[] = new int[crowdingFactor];
		int j=0;
		for(int i=0; i<indivIndex.length; i++){
			int index = rand.irand(currentLength);
			if(j<candidateIndex.length)candidateIndex[j++] = indivIndex[index];
			else break;
			indivIndex[index] =indivIndex[currentLength-1];
			currentLength--;
		}
		return candidateIndex;
	}
	
	//to be written
	void adapt_operators(){}
	void adaptation_stats(){}

}

