
import java.util.*;

abstract class LatticeModelLscape extends ParamLscape{


	int genotype_cardinality = 0;	// for Robbins equilibrium stats

	boolean  inverted = false;   // true ==> return -f as fitness
	StringBuffer seq;
	
	abstract int[][] convertToCoordinates(char[] LMrelativeEncoding);

	/*
	    eval:  computes fitness of a genome for the given landscape
	 */

	float eval(char param[])
	{
		float fitness = 0;

		int coordinates[][] = convertToCoordinates(param);

		for(int i=0; i<seq.length(); i++){
			for(int j=i+1; j<seq.length(); j++){
				int dist = dist(coordinates, i, j);
				if(dist==0){
					fitness = -999; break;
				}else if(dist==1){
					if((j-i)>1){
						if(seq.charAt(i)=='h' && seq.charAt(j)=='h')
							fitness += 1;
					}

				}
			}
		}

		return fitness;   // constraint violation
	}

	boolean legal(char param[]){
		int coordinates[][] = convertToCoordinates(param);

		for(int i=0; i<seq.length(); i++){
			for(int j=i+1; j<seq.length(); j++){
				
				boolean collision=true;
				for(int k=0; k<coordinates[i].length; k++){
					if(coordinates[i][k]!=coordinates[j][k]){
						collision=false;
						break;
					}
				}
				if(collision){
					return false;
				}
				
			}
		}

		return true;
	}
	
	ArrayList<int[]> contactSeperationSignature(char param[], boolean onlyHH)
	{
		ArrayList<int[]> signature = new ArrayList<int[]>(); 

		int coordinates[][] = convertToCoordinates(param);

		for(int i=0; i<seq.length(); i++){
			for(int j=i+3; j<seq.length(); j++){
				
				if(dist(coordinates, i, j)==1){
					//////////////////////////////////////////////////
					int a[] = {i,j};
					if(onlyHH){
						if(seq.charAt(i)=='h' && seq.charAt(j)=='h'){
							signature.add(a);
						}
					}else{
						signature.add(a);
					}
					//////////////////////////////////////////////////
				}
			}
		}

		return signature;   // constraint violation
	}
	
	float avgContactSeperation(char param[])
	{
		int   totalContactSeperation = 0;
		int   n = 0; 

		int coordinates[][] = convertToCoordinates(param);

		for(int i=0; i<seq.length(); i++){
			for(int j=i+1; j<seq.length(); j++){
				int dist = dist(coordinates, i, j);
				if(dist==1){
					if((j-i)>1){
						totalContactSeperation += (j-i);
						n++;
						if((j-i)%2 != 1){
							System.err.println("(j-i)%2 != 1 where i="+i+" j="+j);
							System.exit(-1);
						}
					}

				}
			}
		}

		return (float)totalContactSeperation/n;   // constraint violation
	}
	
	int volume(char param[]){
		int volume = 0;
		
		int coordinates[][] = convertToCoordinates(param);
		
		int dim = coordinates[0].length;
		if(dim!=3){
			System.err.println("dim!=3 where dim="+dim);
			System.exit(-1);
		}
		
		int maxValue[] = new int[dim]; 
		int minValue[] = new int[dim];
		for(int j=0; j<dim; j++){
			maxValue[j] = coordinates[0][j];
			minValue[j] = coordinates[0][j];
		}
		
		for(int i=1; i<coordinates.length; i++){
			for(int j=0; j<dim; j++){
				if(coordinates[i][j]>maxValue[j]){
					maxValue[j] = coordinates[i][j]; 
				}
				if(coordinates[i][j]<minValue[j]){
					minValue[j] = coordinates[i][j]; 
				}
			}
		}
		
		volume = maxValue[0]-minValue[0];
		for(int j=1; j<dim; j++){
			volume *= maxValue[j]-minValue[j];
		}
		
		return volume;
	}

	int dist(int[][] coordinates, int index1, int index2){
		int dist = 0;

		for(int i=0; i<coordinates[index1].length; i++){
			dist += Math.abs(coordinates[index1][i]-coordinates[index2][i]);
		}

		return dist;
	}

	
}
