/********** Modified from IntegerGenome.java ************/

import java.text.DecimalFormat;
import java.util.*;


class RelativeEncodingGenome  extends Genome {
	static RelativeEncodingLscape  lmls;
	public final static char	alphabet[] = {'F', 'L', 'R', 'U', 'D'};

	char    gene[];

	static int        mutation_step_size = 1;
	static boolean    adaptive_mutation = false;
	static float      adaptive_step_size = 0.1F;
	static float      mutation_success_rate;

	/*
   Genome constructor: uses a uniform random number generator to assign
                initial values to each gene from the range specified 
		by the the given lower and upper bounds.
	 */

	RelativeEncodingGenome(EC_rand rand){
		gene = new char[gene_cnt];
		for (int i=0; i<gene_cnt; i++)   gene[i] = alphabet[rand.irand(alphabet.length)];
	}

	RelativeEncodingGenome(){
		gene = new char[gene_cnt];
	}
	RelativeEncodingGenome(char[] a){
		gene = a;
	}
	
	void run_reset(EC_rand rand){
//		do{
//			for (int i=0; i<gene_cnt; i++)   gene[i] = alphabet[rand.irand(alphabet.length)];
//		}while(!this.legal());
		for (int i=0; i<gene_cnt; i++)   gene[i] = 'F';
	}

	/*
   IntegerGenomeClassInit: initializes the genome/phenome mapping for entire class
	 */

	static void RelativeEncodingGenomeClassInit(Properties props, Landscape lscape, EA ea) 
	{

		if (!lscape.lscape_type.equals("RelativeEncoding"))
			Util.ec_error("Can't map " + lscape.lscape_type + " into relative encoding representation");

		lmls = (RelativeEncodingLscape) lscape;

		gene_cnt = lmls.param_cnt;

		System.out.println("  Using a genome with " + gene_cnt + " relative encoding(char) gene(s).");

		genotype_cardinality = lmls.genotype_cardinality;

		ea.global_max = new RelativeEncodingIndividual();
		ea.global_min = new RelativeEncodingIndividual();
		ea.global_max.fitness = Float.NEGATIVE_INFINITY;
		ea.global_min.fitness = Float.POSITIVE_INFINITY;
		ea.parent_pop = new RelativeEncodingPopulation(ea, ea.parent_pop_size, ea.rand);

		GenomeClassInit(props, ea);

		System.out.println("Relative Encoding: Mutation type using default.");
		String mutation_spec = "random";
		mutation_type = random_mutation;
		mutation_probability = 1.0F / Genome.gene_cnt;

		String mutation_options = props.getProperty("Mutation");
		if(mutation_options != null){
			StringTokenizer st = new StringTokenizer(mutation_options," ");
			st.nextToken();
			if(st.hasMoreTokens()){
				mutation_probability = Float.parseFloat(st.nextToken());
			}
		}

	}




	/*
  mutate:  mutates the given genome
	 */

	void mutate(EA ea)
	{
		switch (mutation_type) {

		case random_mutation:
			random_mutate(ea.rand);
			break;
		default:
			Util.ec_error("illegal Relative Encoding mutation type: " + mutation_type);
		}
	}

	void random_mutate(EC_rand rand)
	{
		char new_val[] = new char[gene_cnt];
		int mutated = 0;

		for (int i=0; i<gene_cnt; i++)
		{
			if (rand.frand() <= mutation_probability)
			{     
				new_val[i] = alphabet[rand.irand(alphabet.length)];
				mutation_step_sum += 1;
				mutated++;
			}
			else  new_val[i] = gene[i];
		}
		
		/* either ********/
//		if (mutated>0)gene = new_val;
		/* or ************/
		if (mutated>0)
		{
			if (lmls.legal(new_val))
				gene = new_val;
			else{
				invalid_mutation_count++;
			}				
		}

		mutation_count += mutated;
		return;
	}
	
	void absoluteEncoding_random_mutate(EC_rand rand)
	{
		char new_val[] = new char[gene_cnt];
		int mutated = 0;

		for (int i=0; i<gene_cnt; i++)
		{
			if (rand.frand() <= mutation_probability)
			{     
				new_val[i] = alphabet[rand.irand(alphabet.length)];
				
				mutated++;
				mutation_step_sum += 1;
			}
			else  new_val[i] = gene[i];
		}
		
		/* either ********/
//		if (mutated>0)gene = new_val;
		/* or ************/
		if (mutated>0)
		{
			if (lmls.legal(new_val))
				gene = new_val;
			else{
				invalid_mutation_count++;
			}				
		}

		mutation_count += mutated;
		return;
	}
	
	

	/*
  crossover:
	 */

	void crossover(Genome mate, EC_rand rand)
	{
//		int count=0;
//		do{
			switch (crossover_type) {
	
			case one_point:
				one_pt_crossover((RelativeEncodingGenome)mate, rand);
				break;
	
			case two_point:
				two_pt_crossover((RelativeEncodingGenome)mate, rand);
				break;
	
			case uniform:
				uniform_crossover((RelativeEncodingGenome)mate, rand);
				break;
	
			default:
				Util.ec_error("illegal Integer crossover type: " + crossover_type);
			}
//			count++;
//			if(count>1000){
//				System.err.println("Oh, my god. Forget it");
//				break;
//			}
//		}while( ! lmls.legal(this.gene));
	}


	/*
   one_pt_crossover:  copies selected genes from mate into this genome
	 */

	void one_pt_crossover(RelativeEncodingGenome mate, EC_rand rand)
	{
		int cut_point, i;
		char tmp;

		if (gene_cnt == 1)  return;    // need at least two genes

		if (keep_both_sides) {
			if (gene_cnt == 2) {	//easy case
				tmp = this.gene[1];
				this.gene[1] = mate.gene[1];
				mate.gene[1] = tmp;
			}
			else {
				cut_point = rand.irand(gene_cnt-1)+1;
				for (i=cut_point; i<gene_cnt; i++)  {
					tmp = this.gene[i];
					this.gene[i] = mate.gene[i];
					mate.gene[i] = tmp;
				}
			}
		}
		else {
			if (gene_cnt == 2)  this.gene[1] = mate.gene[1];   // easy case
			else {
				cut_point = rand.irand(gene_cnt-2);
				for (i=cut_point+1; i<gene_cnt; i++)  this.gene[i] = mate.gene[i];
			}
		}

		return;
	}


	/*
   two_pt_crossover:  copies selected genes from mate into this genome
	 */

	void two_pt_crossover(RelativeEncodingGenome mate, EC_rand rand)
	{
		int cut_point1, cut_point2, i;
		char tmp;

		if ((gene_cnt == 1) || (gene_cnt ==2))  return;  // need at least 3 genes

		if (keep_both_sides) {
			if (gene_cnt == 3) {  // easy case
				tmp = this.gene[1];
				this.gene[1] = mate.gene[1];
				mate.gene[1] = tmp;
			}
			else
			{
				cut_point1 = rand.irand(gene_cnt);
				cut_point2 = rand.irand(gene_cnt);
				while (cut_point2 == cut_point1) 
					cut_point2 = rand.irand(gene_cnt);  // want distinct cut points

				if (cut_point1 < cut_point2)
					for (i=cut_point1+1; i<=cut_point2; i++)  {
						tmp = this.gene[i];
						this.gene[i] = mate.gene[i];
						mate.gene[i] = tmp;
					}
				else
					for (i=cut_point2+1; i<=cut_point1; i++)  {
						tmp = this.gene[i];
						this.gene[i] = mate.gene[i];
						mate.gene[i] = tmp;
					}
			}
		}
		else {
			if (gene_cnt == 3)  this.gene[1] = mate.gene[1];  // easy case
			else
			{
				cut_point1 = rand.irand(gene_cnt);
				cut_point2 = rand.irand(gene_cnt);
				while (cut_point2 == cut_point1) 
					cut_point2 = rand.irand(gene_cnt);  // want distinct cut points

				if (cut_point1 < cut_point2)
					for (i=cut_point1+1; i<=cut_point2; i++)  this.gene[i] = mate.gene[i];
				else
					for (i=cut_point2+1; i<=cut_point1; i++)  this.gene[i] = mate.gene[i];
			}
		}

		return;
	}






	/*
   uniform_crossover:  copies genes from mate into this genome probabilistically
	 */

	void uniform_crossover(RelativeEncodingGenome mate, EC_rand rand)
	{
		int i; char tmp;

		if (keep_both_sides) {
			for (i=0; i<gene_cnt; i++)  
				if (rand.frand() <= uniform_crossover_probability) {
					tmp = this.gene[i];  
					this.gene[i] = mate.gene[i];
					mate.gene[i] = tmp;
				}
		}
		else {
			for (i=0; i<gene_cnt; i++)  
				if (rand.frand() <= uniform_crossover_probability)  
					this.gene[i] = mate.gene[i];
		}

		return;
	}



	/*
   global_stats:  print global operator stats
	 */

	void global_stats()
	{
		System.out.println(" Total number of mutations: " + mutation_count);
		System.out.println(" Number of constraint violations: " + 
				invalid_mutation_count);
		System.out.println(" Average mutation step size: " +
				mutation_step_sum/mutation_count);
	}



	/*
   eval:
	 */

	float eval()
	{
		return (lmls.eval(gene));
	}

	


	DecimalFormat i6 = new DecimalFormat("######0");


	/*
   print_genome:  prints the gene values as real numbers
	 */

	void print()
	{
		int  i;

		for (i=0; i<gene_cnt; i++)
			System.out.print(gene[i]);
	}



	Genome copy()
	{
		RelativeEncodingGenome  dupl = new RelativeEncodingGenome();

		dupl.gene_cnt = gene_cnt;
		dupl.gene = new char[gene_cnt];

		System.arraycopy(gene, 0, dupl.gene, 0, gene_cnt);

		return ((Genome) dupl);
	}



	/*
  distance:  computes Hamming distance between self and specified partner
	 */

	final float distance(Genome partner)
	{
		int diff=0;

		for (int i=0; i<Genome.gene_cnt; i++) { 
			if(this.gene[i] != ((RelativeEncodingGenome)partner).gene[i]){
				diff++;
			}
		}

		return(diff);
	}
	
	@Override
	Genome createMirrorGenome(){
		
		char[] newGene = new char[gene.length];
		for(int i=0; i<gene.length; i++){
			switch(gene[i]){
				case 'F':
					newGene[i] = 'F';
					break;
				case 'R':
					newGene[i] = 'L';
					break;
				case 'L':
					newGene[i] = 'R';
					break;
				case 'U':
					newGene[i] = 'D';
					break;
				case 'D':
					newGene[i] = 'U';
					break;
				default:
					System.err.println("not FRLUD");
					System.exit(-1);
			}
		}
		
		return new RelativeEncodingGenome(newGene);
	}
	
	public boolean legal(){
		return lmls.legal(gene);
	}
	
	public float avgContactSeperation(){
		return (lmls.avgContactSeperation(gene));
	}
	
	public float volume(){
		return (lmls.volume(gene));
	}
	
	public ArrayList<int[]> contactSeperationSignature(boolean onlyHH){
		return (lmls.contactSeperationSignature(gene, onlyHH));
	}
	
	public final void absoluteMutate(int index, char replacedOne){
		if(index>=gene.length)return;
		char oldGene[] = gene.clone();
		absoluteMutateOne(gene,index,replacedOne);
		if(!legal()){
			gene = oldGene;
		}
	}
	
	private final void absoluteMutateOne(char[] seq, int index, char replacedOne){
		
		if(index>=(seq.length-1)){
			seq[index] = replacedOne;
			return;
		}
		
		// Input : seq, index
		// Output: absSeq,replacedAbsSeqChr
		char absSeq[] = new char[index+2+1];
		char replacedAbsSeqChr = '?'; //compiler happy
		absSeq[0] = 'X';
//		for(int i=0; i<seq.length; i++){
		for(int i=0; i<=index+1; i++){
			String cmdList = getCmdlistFromAbsoluteChar(absSeq[i]);
			absSeq[i+1] = cmdList.charAt(chrToCmdlistIndex(seq[i]));
			if(i==index)replacedAbsSeqChr = cmdList.charAt(chrToCmdlistIndex(replacedOne));
		}
		
		// Input : replacedAbsSeqChr, absSeq[index+1+1]
		// Output: seq[index+1]
		String newCmdList = getCmdlistFromAbsoluteChar(replacedAbsSeqChr);
		int j=0;
		for(j=0; j<newCmdList.length(); j++){
			if(newCmdList.charAt(j) == absSeq[index+1+1])break;
		}
		if(j==newCmdList.length())return;
		
		seq[index]   = replacedOne;
		seq[index+1] = cmdlistIndexToChr(j);
	}
	
	private final String getCmdlistFromAbsoluteChar(char AbsoluteChr){
		switch (AbsoluteChr){
		case 'X':
			return RelativeEncodingLscape.deltaX_cmdList;
		case 'x':
			return RelativeEncodingLscape.deltaXN_cmdList;
		case 'Y':
			return RelativeEncodingLscape.deltaY_cmdList;
		case 'y':
			return RelativeEncodingLscape.deltaYN_cmdList;
		case 'Z':
			return RelativeEncodingLscape.deltaZ_cmdList;
		case 'z':
			return RelativeEncodingLscape.deltaZN_cmdList;
		default:
			System.err.println("absSeq[i] != legal");
			System.exit(-1);
			return "";
		}
	}
	
	// X, x, Y, y, Z, z
	private final int chrToCmdlistIndex(char chr){
		if(chr == 'F'){
			return 0;
		}else if(chr == 'R'){
			return 1;
		}else if(chr == 'L'){
			return 2;
		}else if(chr == 'U'){
			return 3;
		}else if(chr == 'D'){
			return 4;
		}else{
			System.out.println("You got an alien relative move character! Have a nice day!");
			System.exit(-1);
			return -1; //compiler happy
		}
	}
	
	private final char cmdlistIndexToChr(int index){
		if(index == 0){
			return 'F';
		}else if(index == 1){
			return 'R';
		}else if(index == 2){
			return 'L';
		}else if(index == 3){
			return 'U';
		}else if(index == 4){
			return 'D';
		}else{
			System.out.println("You got an alien relative move index! Have a nice day!");
			System.exit(-1);
			return '-'; //compiler happy
		}
	}
	
}





