import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;


public class SCGA extends EA{
	
	List<Individual> a;
	double speciesDistance;
	SCGA(Properties props, Landscape lscape, int rand_seed){
	    super(rand_seed);
	    ea_type = "SCGA";
	    EA_init(props, lscape);         // finish general EA initialization;
	    
	    StringTokenizer st = new StringTokenizer(props.getProperty("EA")," ");
	    st.nextToken();
	    if(st.hasMoreTokens()){
	    	speciesDistance = Double.parseDouble(st.nextToken());
	    }else{
	    	speciesDistance = 5;
	    }
	    
	    System.out.println("SCGA: It uses species distance="+speciesDistance);
	    
	    
    }
	
	/*
	   run: the specified EA for the specified time
	*/
	
	void run(int birth_limit)
	{
		birth_cnt = 0;
		generation_cnt = 0;
		NFE = 0;
		global_max.fitness = Float.NEGATIVE_INFINITY;
		global_min.fitness = Float.POSITIVE_INFINITY;
		parent_pop.run_reset();
		dump();
		
//		FitnessComparator c = new FitnessComparator();
//		Arrays.sort(parent_pop.indiv, c);
//		
		if(speciesDistance==0 ){
			Util.ec_error("speciesDistance == 0 !"); System.exit(-1);
		}
		while ( (birth_cnt < birth_limit) & (NFE <= EC4.maxNFE) ) 
		{
			if (report_fixed_gen || report_fixed_fit)  if (parent_pop.fixed_point()) {
				break;
			}
			
			List<Individual> a = speciesSeed(speciesDistance);

			offspring_pop = select_and_clone_parents();

			introduce_offspring_variation();

			evaluate_offspring();
			
			select_survivors();

			generation_cnt ++;
			
			conserveSpecies(speciesDistance, a);
			
			if(report_pop_dump)speciesSeedDump(a);
			
			dump();
//			adapt_operators();
//			adaptation_stats();
		}
	}
	
	final List<Individual> speciesSeed(double speciesDistance){
		
		List<Individual> seedList = new ArrayList<Individual>();
		FitnessComparator c = new FitnessComparator();  //should have been sorted by conserveSpecies()
		Arrays.sort(parent_pop.indiv, c);
		//System.out.println("*-Sorted list--------------------");
		//dump();
		//System.out.println("*-Species seeds-----------------");
		seedList.add(parent_pop.indiv[0]);
		for(int i=1; i<parent_pop.indiv.length; i++){
			boolean found = false;
			for(int j=0; j<seedList.size(); j++){
				if(parent_pop.indiv[i].genome.distance(seedList.get(j).genome)<speciesDistance){
					found = true;
					break;
				}
			}
			if(!found){
				seedList.add(parent_pop.indiv[i]);
			}

		}
		return seedList;
	}
	
	// to be validated
	final void conserveSpecies(double speciesDistance, List<Individual> seedList){
		FitnessComparator c = new FitnessComparator();
		Arrays.sort(parent_pop.indiv, c);
		
		boolean marked[] = new boolean[parent_pop.pop_size];
		boolean inserted[] = new boolean[seedList.size()];
		
		for(int i=0; i<parent_pop.pop_size; i++)marked[i]=false;
		for(int i=0; i<seedList.size(); i++)inserted[i]=false;
		
		for(int j=0; j<seedList.size(); j++){
			//find species offspring, select the worst un-marked one.
			for(int i=parent_pop.indiv.length-1; i>=0; i--){
				if(!inserted[j] & !marked[i]){
					if (parent_pop.indiv[i].genome.distance(seedList.get(j).genome) < speciesDistance){
						parent_pop.indiv[i] = seedList.get(j);
						marked[i]=true;
						inserted[j]=true;
						break;
					}
				}
			}
			
			//no species offspring, select the worst un-marked one (not same species).
			if(!inserted[j]){
				for(int i=parent_pop.indiv.length-1; i>=0; i--){
					if(!marked[i]){
						parent_pop.indiv[i] = seedList.get(j);
						marked[i]=true;
						inserted[j]=true;
						break;
					}
				}
			}
		}
		
	}


	//Dump an arrayList<Individual>
	final void speciesSeedDump(List<Individual> seedList){
	    System.out.println(" Indiv  birthdate   fitness   fCount   seedSurv      gene values"); 
	
	    int i=0;
	    for(Individual speciesSeed : seedList){
	      System.out.print(Util.pad_left(int4.format(i+1), 5) + 
		  Util.pad_left(int4.format(speciesSeed.birthdate), 7) + 
		  Util.pad_left(f12_5.format(speciesSeed.fitness), 15) + 
		  "  ");
	      speciesSeed.genome.print();
	      System.out.println("");
	      i++;
	    }
	}
	
	//to be written
	void adapt_operators(){}
	void adaptation_stats(){}
}
