import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

/*       Copyrighted by Kenneth De Jong - 2005          */
/* Licensed under the Academic Free License version 2.1 */

/*
  Util:  miscellaneous EC utility routines
*/


class Util  {


/*
  pad_left:  kludge until formatted I/O gets friendlier
*/


    static String blanks = new String("                                                   ");
    static String padded;

    static String pad_left(String str, int width) {
	if (width > blanks.length())  return (str);
	padded = blanks.substring(0, width - str.length());
	return (padded.concat(str));
        }


/*
  Standard EC messages:
*/


    static void ec_warn(String msg) {
	System.out.println(" ** EC warning: " + msg);
        }

    static void ec_error(String msg) {
	System.out.println(" ** EC error: " + msg);
	System.exit(1);
        }
    
    /**
     * Calculates the standard deviation of an array
     * of numbers.
     * see Knuth's The Art Of Computer Programming
     * Volume II: Seminumerical Algorithms
     * This algorithm is slower, but more resistant to error propagation.
     *
     * @param data Numbers to compute the standard deviation of.
     * Array must contain two or more numbers.
     * @return standard deviation estimate of population
     * ( to get estimate of sample, use n instead of n-1 in last line )
     */
    public static double SDKnuth ( double[] data )
    {
    	final int n = data.length;
    	if ( n < 2 )
    	{
    		return Double.NaN;
    	}
    	double avg = data[0];
    	double sum = 0;
    	for ( int i = 1; i < data.length; i++ )
    	{
    		double newavg = avg + ( data[i] - avg ) / ( i + 1 );
    		sum += ( data[i] - avg ) * ( data [i] -newavg ) ;
    		avg = newavg;
    	}
    	return Math.sqrt( sum / ( n - 1 ) );
    }
    
    //================================================= mean
    public static double mean(double[] p) {
        double sum = 0;  // sum of all the elements
        for (int i=0; i<p.length; i++) {
            sum += p[i];
        }
        return sum / p.length;
    }//end method mean
    
    // ================================================== median
	//  Precondition: Array must be sorted
	public static double median(double[] m) {
	   int middle = m.length/2;  // subscript of middle element
	   if (m.length%2 == 1) {
	       // Odd number of elements -- return the middle one.
	       return m[middle];
	   } else {
	      // Even number -- return average of middle two
	      // Must cast the numbers to double before dividing.
	      return (m[middle-1] + m[middle]) / 2.0;
	   }
	}//end method median
	
	//	===================================================== max
	public static double max(double[] t) {
	    double maximum = t[0];   // start with the first value
	    for (int i=1; i<t.length; i++) {
	        if (t[i] > maximum) {
	            maximum = t[i];   // new maximum
	        }
	    }
	    return maximum;
	}//end method min
	public static double min(double[] t) {
	    double minimum = t[0];   // start with the first value
	    for (int i=1; i<t.length; i++) {
	        if (t[i] < minimum) {
	            minimum = t[i];   // new maximum
	        }
	    }
	    return minimum;
	}//end method min
    
    /*********Print current time************/
    public static final String DATE_FORMAT_NOW = "yyyy-MM-dd HH:mm:ss";
    public static String now() {
      Calendar cal = Calendar.getInstance();
      SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
      return sdf.format(cal.getTime());

    }
    
    /**ArrayList of Double to double[]*/
	public static double[] arrayListOfDoubleToArray(ArrayList dbl){
		int num = dbl.size();
		double[] d = new double[num];
		for (int i=0; i<num; i++)d[i]= ((Double)dbl.get(i)).doubleValue();
		return d;
	}
	

}
